/* main.c - this file is part of DeSmuME
 *
 * Copyright (C) 2006,2007 DeSmuME Team
 * Copyright (C) 2007 Pascal Giard (evilynux)
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <stdio.h>
#define printf pspDebugScreenPrintf

PSP_MODULE_INFO("DeSmuME", 0, 1, 1);

#include "../MMU.h"
#include "../NDSSystem.h"
#include "../cflash.h"
#include "../debug.h"
#include "../sndsdl.h"
#include "../ctrlssdl.h"
#include "../render3D.h"
#include "../gdbstub.h"

volatile BOOL execute = FALSE;

static float nds_screen_size_ratio = 1.0f;

#define DISPLAY_FPS 1

#ifdef DISPLAY_FPS
#define NUM_FRAMES_TO_TIME 15
#endif

#define FPS_LIMITER_FRAME_PERIOD 8

static SDL_Surface * surface;

/* Flags to pass to SDL_SetVideoMode */
static int sdl_videoFlags = 0;

SoundInterface_struct *SNDCoreList[] = {
  &SNDDummy,
  &SNDFile,
  &SNDSDL,
  NULL
};

GPU3DInterface *core3DList[] = {
&gpu3DNull
};

struct my_config {
  u16 arm9_gdb_port;
  u16 arm7_gdb_port;

  int disable_sound;
  int disable_limiter;

  const char *nds_file;
  const char *cflash_disk_image_file;
};

static void
init_config( struct my_config *config) {
  config->disable_sound = 0;
  config->disable_limiter = 1;
#ifdef __psp__
  config->nds_file = "test.nds";
#else
  config->nds_file = NULL;
#endif
  config->cflash_disk_image_file = NULL;
}

static void
Draw( void) {
  SDL_Surface *rawImage;
	
  rawImage = SDL_CreateRGBSurfaceFrom((void*)&GPU_screen, 256, 384, 16, 512, 0x001F, 0x03E0, 0x7C00, 0);
  if(rawImage == NULL) return;
	
  SDL_BlitSurface(rawImage, 0, surface, 0);

  SDL_UpdateRect(surface, 0, 0, 0, 0);
  
  SDL_FreeSurface(rawImage);
  return;
}

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();

	return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

int main(int argc, char *argv[])
{
  static unsigned short keypad = 0;
  struct my_config my_config;
  u32 last_cycle = 0;
  struct armcpu_memory_iface *arm9_memio = &arm9_base_memory_iface;
  struct armcpu_memory_iface *arm7_memio = &arm7_base_memory_iface;
  struct armcpu_ctrl_iface *arm9_ctrl_iface;
  struct armcpu_ctrl_iface *arm7_ctrl_iface;
  #ifndef __psp__
  int limiter_frame_counter = 0;
  SDL_sem *fps_limiter_semaphore;
  SDL_TimerID limiter_timer;
  #endif
  int sdl_quit = 0;
  
  pspDebugScreenInit();
  scePowerSetClockFrequency(333, 333, 166);
  SetupCallbacks();
  
#ifdef DISPLAY_FPS
  u32 fps_timing = 0;
  u32 fps_frame_counter = 0;
  u32 fps_previous_time = 0;
  u32 fps_temp_time;
#endif

  /* this holds some info about our display */
  const SDL_VideoInfo *videoInfo;

  init_config( &my_config);
  
  SceCtrlData pad;
  pspDebugScreenPrintf("DeSmuME PSP Useless Edition\n\n");
  pspDebugScreenPrintf("Push X to boot test.nds without sound\n");
  pspDebugScreenPrintf("Push O to boot test.nds with sound");
  
  while (1) {
		sceCtrlReadBufferPositive(&pad, 1);
		if (pad.Buttons & PSP_CTRL_CROSS){
			my_config.disable_sound = 1;
			break;
		}
		if (pad.Buttons & PSP_CTRL_CIRCLE){
			my_config.disable_sound = 0;
			break;
		}
	}
  
  pspDebugScreenClear(); 

  NDS_Init( arm9_memio, &arm9_ctrl_iface,
            arm7_memio, &arm7_ctrl_iface);
  
  if ( !my_config.disable_sound) {
    SPU_ChangeSoundCore(SNDCORE_SDL, 735 * 4);
  }

  if (NDS_LoadROM( my_config.nds_file, MC_TYPE_AUTODETECT, 1, my_config.cflash_disk_image_file) < 0) {
    SceCtrlData pad;
	printf("Error loading %s\n", my_config.nds_file);
	sceKernelDelayThread(100000);
	do { sceCtrlReadBufferPositive(&pad, 1);
	} while (pad.Buttons == 0);
	sceKernelExitGame();
  }

  execute = TRUE;
  
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1)
    {
      fprintf(stderr, "Error trying to initialize SDL: %s\n",
              SDL_GetError());
      return 1;
    }
  
  SDL_WM_SetCaption("Desmume SDL", NULL);

  /* Fetch the video info */
  videoInfo = SDL_GetVideoInfo( );
  if ( !videoInfo ) {
    fprintf( stderr, "Video query failed: %s\n", SDL_GetError( ) );
    exit( -1);
  }

  /* This checks if hardware blits can be done */
  if ( videoInfo->blit_hw )
    sdl_videoFlags |= SDL_HWACCEL;
    sdl_videoFlags |= SDL_SWSURFACE;
    surface = SDL_SetVideoMode(256, 384, 32, sdl_videoFlags);

    if ( !surface ) {
      fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError( ) );
      exit( -1);
    }
  
  #ifndef __psp__
  /* Initialize joysticks */
  if(!init_joy()) return 1;
  /* Load our own keyboard configuration */
  #endif

  while(!sdl_quit) {

    // Look for queued events and update keypad status
    sdl_quit = process_ctrls_events( &keypad, NULL, nds_screen_size_ratio);
    
	
    // Update mouse position and click
    if(mouse.down) {
		NDS_setTouchPos(mouse.x, mouse.y);
	}
	
    if(mouse.click)
      { 
        NDS_releasTouch();
        mouse.click = FALSE;
      }
	
	update_keypad(keypad);     /* Update keypad */
	last_cycle = NDS_exec((560190 << 1) - last_cycle, FALSE);
	
	if ( !my_config.disable_sound) {
		SPU_Emulate();
	}

	Draw();

#ifdef DISPLAY_FPS
    fps_frame_counter += 1;
    fps_temp_time = SDL_GetTicks();
    fps_timing += fps_temp_time - fps_previous_time;
    fps_previous_time = fps_temp_time;

    if ( fps_frame_counter == NUM_FRAMES_TO_TIME) {
	  #ifndef __psp__
      char win_title[100];
	  #endif
      float fps = (float)fps_timing;
      fps /= NUM_FRAMES_TO_TIME * 1000.f;
      fps = 1.0f / fps;
	  pspDebugScreenSetXY(0,0);
      printf("FPS %f", fps);
      fps_frame_counter = 0;
      fps_timing = 0;
	  #ifndef __psp__
      sprintf( win_title, "Desmume %f", fps);
      SDL_WM_SetCaption( win_title, NULL);
	  #endif
    }
#endif
  }


  #ifndef __psp__
  /* Unload joystick */
  uninit_joy();

  if ( !my_config.disable_limiter) {
    /* tidy up the FPS limiter timer and semaphore */
    SDL_RemoveTimer( limiter_timer);
    SDL_DestroySemaphore( fps_limiter_semaphore);
  }
  #endif
  
  SDL_Quit();
  NDS_DeInit();

  return 0;
}
